-- All code written and maintained by Yewchi 
-- zyewchi@gmail.com

-- TODO BBC spell links not auto cancel

FLEXI = {}

FLEXI.DEBUG = false
FLEXI.VERBOSE = FLEXI.DEBUG and true

local TEST_FORCE_SEND_FLEXI_CODE = false

FLEXI.BUILD_VERSION = select(4, GetBuildInfo())
FLEXI.WOW_VC = FLEXI.BUILD_VERSION < 20000
FLEXI.WOW_BCC = FLEXI.BUILD_VERSION > 20000 and FLEXI.BUILD_VERSION < 30000

--- Initialization --
--
------- Flags --
--
local FLEXI_CODE_IS_ITEM = 0xF0
local FLEXI_CODE_IS_SPELL = 0xF1

local FLEXI_CHAT_CMD_CHANNEL_NOT_FOUND = 0xF0
local FLEXI_CHAT_CMD_LINKABLE_NOT_FOUND = 0xF1

local oldSubFunc = SubstituteChatMessageBeforeSend
local slappedForErrorsOrPoorReturnValues = false
--
------ /Flags --

------- Table key references --
--
FLEXI.IS_ID = 1
FLEXI.IS_ENCHANT_ID = 2
FLEXI.IS_GEM_ID1 = 3
FLEXI.IS_GEM_ID2 = 4
FLEXI.IS_GEM_ID3 = 5
FLEXI.IS_GEM_ID4 = 6
FLEXI.IS_SUFFIX_ID = 7
FLEXI.IS_UNIQUE_ID = 8
FLEXI.IS_LINK_LEVEL = 9
FLEXI.IS_SPEC_ID = 10
FLEXI.IS_UPGRADE_TYPE_ID = 11
FLEXI.IS_DIFFICULTY_ID = 12

FLEXI.SS_ID = 1

FLEXI.SPELL_INFO_NAME = 1
FLEXI.SPELL_INFO_RANK = 2
FLEXI.SPELL_INFO_ID = 7

local START_STRING = 1
local STOP_STRING = 2
--
------ /Table key references --

------- Static vals --
--
FLEXI.flexiSpecifier = "|cff18D050[flexi]".."|r "
FLEXI.debugSpecifier = "|cffC020C0<DEBUG>".."|r: "
FLEXI.errSpecifier = "|cffC01050<ERR>".."|r: "
FLEXI.infoSpecifier = "|cff7070FF<INFO>".."|r: "

FLEXI.SYSTEM_COLOUR = "|cffFFCC44"

FLEXI.INFO_SYNTAX = "|r/{channel to post message to} {message with Shift-Clicked [Item Link] or Spell Name preceeded by a '$'}"
FLEXI.INFO_EXAMPLE = FLEXI.SYSTEM_COLOUR.."Example flexi for chat channel 4:\n|cffFFFFFF/4 Selling $[Leather Belt] 2 gold!\n"..FLEXI.SYSTEM_COLOUR.."Example flexi for guild chat:\n|cffFFFFFF/g Who needs $Enchanting?|r"
FLEXI.CHANNEL_NOT_RECOGNIZED = FLEXI.flexiSpecifier..FLEXI.SYSTEM_COLOUR.."Nope, |cffFF3311channel not recognized! "..FLEXI.SYSTEM_COLOUR.."Type '/flexi help', or try again."
FLEXI.INFO_HELPTEXT = FLEXI.flexiSpecifier..FLEXI.infoSpecifier..FLEXI.SYSTEM_COLOUR.."Usage : Type into chat...\n"..FLEXI.INFO_SYNTAX.."\n"..FLEXI.INFO_EXAMPLE

local cmdPrefixAuto = "$"
local cmdPrefixConsole = ""

local AMMEND_FORMAT_STRING_NO_SPACE = "%s%s"
local AMMEND_FORMAT_STRING_SPACE = "%s %s"

local IMPOSSIBLE_LINK_CHANNEL_START = "["
local IMPOSSIBLE_LINK_CHANNEL_STOP = "]"
local IMPOSSIBLE_LINK_CHANNEL_TBL = {IMPOSSIBLE_LINK_CHANNEL_START, IMPOSSIBLE_LINK_CHANNEL_STOP}
local POSSIBLE_LINK_CHANNEL_START = ""
local POSSIBLE_LINK_CHANNEL_STOP = ""
local POSSIBLE_LINK_CHANNEL_TBL = {POSSIBLE_LINK_CHANNEL_START, POSSIBLE_LINK_CHANNEL_STOP}

local MAX_CHAT_MESSAGE_LEN = 256

local NUM_OF_ITEM_SPLITS = 13
local NUM_OF_SPELL_SPLITS = 3
--
------- /Static vals --

------- Return values --
--
local MSG_INTEGRITY_INVALID_ESCAPE_CHARACTER 	= 0xFF
local MSG_INTEGRITY_WORD_LOST 					= 0xFE
--
------ /Return values --

------- AddOn Variables --
--
local Frame_ChatMessage
local Frame_SpellBook
--
------ /AddOn Variables --
--
-- /Initialization --

--- TODO: Funcs / files are not sorted in a very modular way. Need better file-scope definitions for extendability.
--- Funcs --
--
-------------- SplitString()
local function SplitString(str, splitter)
	if str == nil or splitter == nil then
		error("FLEXI: Invalid SplitString args:: SplitString(str = "..(str == nil and "nil" or str)..", splitter = "..(splitter == nil and "nil" or splitter)..")")
	end
	local str_arr = {}
	local i = 1
	--print(str)
	saveit2=str
	for sub in str:gmatch(splitter) do
		str_arr[i] = sub
		i = i + 1
	end
	--print("SS Returns "..FLEXI.PrintableTable(str_arr))
	return str_arr
end

-------------- GetCmdPrefix()
local function GetCmdPrefix(cmdType)
	if cmdType == "auto" then
		return cmdPrefixAuto
	elseif cmdType == "console" then
		return cmdPrefixConsole
	end
	return nil
end

-------------- GetLinkStartAndStopString()
local function GetLinkStartAndStopString(channel, linkType)
if FLEXI.DEBUG and TEST_FORCE_SEND_FLEXI_CODE then return IMPOSSIBLE_LINK_CHANNEL_TBL end
	if type(channel) == "number" then
		local _, channelName = GetChannelName(channel)
		if linkType == "item" and not (channelName:match("General") or channelName:match("LocalDefense") or channelName:match("LookingForGroup")) then
			return POSSIBLE_LINK_CHANNEL_TBL
		end
	elseif linkType == "item" then -- Was not one of the "/#" no-item link channels, i.e. it may be /s, /g, /w, /i, etc.
		return POSSIBLE_LINK_CHANNEL_TBL
	end
	return IMPOSSIBLE_LINK_CHANNEL_TBL
end

-------------- IsMessageIntegrityMaintained()
local function IsMessageIntegrityMaintained(flexifiedMsg, originalMsg)
	local originalMsgLen = originalMsg:len()

	local flexifiedSpecialsSplit = SplitString(
			FLEXI.SanitizeSpecialCharacters(
				flexifiedMsg:gsub("flexi.+", ""):gsub("FLEXI.+", ""):gsub("%[(.+)%]", ""), -- TODO: Players are very unlikely but able to enclose some of their message in "[", "]".
				"[^%$%[%]%s]+"
			)
	)
if FLEXI.DEBUG then print(FLEXI.flexiSpecifier..FLEXI.debugSpecifier.."Comparing flexified to original:\n"..FLEXI.Printable(flexifiedMsg).."\n"..FLEXI.Printable(originalMsg)) end	
	if flexifiedMsg:find(FLEXI.SanitizeSpecialCharacters(originalMsg)) and not string.lower(flexifiedMsg):find("flexi") then
		return true -- Message was unchanged. N.B.: This gives requirement for InterpretMessageSendToChannel() to never edit messages to channels which could post the item link.
	end

	local lastSplit = 1
	for i=1,#flexifiedSpecialsSplit,1 do
		local thisSplit = flexifiedSpecialsSplit[i]
		local foundAt, foundAtEnd = originalMsg:sub(lastSplit, originalMsgLen):find(thisSplit)
if FLEXI.VERBOSE then print(FLEXI.flexiSpecifier..FLEXI.debugSpecifier.."Checking split: '"..FLEXI.Printable(thisSplit).."'. foundAt="..FLEXI.Printable(foundAt)..", foundAtEnd="..FLEXI.Printable(foundAtEnd)..".") end
		if type(foundAt) == "number" then
			lastSplit = foundAtEnd
		else
			return MSG_INTEGRITY_WORD_LOST
		end
	end
	
	return true -- Success
end

--- Takes the args of the flexi chat command, channelNumber, the link type:['item', 'spell'], and the current loop index of InterpretMessageSendToChannel()
--- Returns the string to add to chat for this link, itemLinkImpossible boolean, and the index at which i should continue
--- TODO: When spells links are allowed to be entered into the chat frame (which version will Blizz do it?), spells need to be checked in this function.
-------------- SanitizeLinkGetRawValues()
local function SanitizeLinkGetRawValues(args, channelNumber, checkLinkType, linkEscapePrefix, currentIndex)
	local i = currentIndex
	local thisWord = args[i]
	local isItem = checkLinkType == "item"

	local preLinkWord = isItem and thisWord:match("(.+)"..linkEscapePrefix.."|c........|Hitem") or thisWord:match("(.+)"..linkEscapePrefix.."|c........|Hspell")
	preLinkWord = (preLinkWord ~= nil and preLinkWord or "")
	
	local linksEnd = isItem and thisWord:match(linkEscapePrefix.."|c........|Hitem.-%[.-%]|h|r(.-)$") or thisWord:match(linkEscapePrefix.."|c........|Hspell.-%[.-%]|h|r(.-)$")
	
	-- TODO Shouldn't be declared here
	local linkStartAndStopString = GetLinkStartAndStopString(channelNumber, checkLinkType)
	local linkImpossible = linkStartAndStopString == IMPOSSIBLE_LINK_CHANNEL_TBL
	
	local linkName
	
	if linkImpossible then
		-- First word of the link, with preceeding characters "20x[Mageweave..."
		local fullFind = thisWord:match(linkEscapePrefix..".-|h%[(.-)%]|h|r")
		linkName = string.format("%s%s%s", preLinkWord, linkStartAndStopString[START_STRING], (fullFind ~= nil and fullFind ~= "") and fullFind or thisWord:match(linkEscapePrefix..".-|h%[(.+)")) -- ResidentSleeper

		-- Add each following word of the name of this link "... Cloth..."
		i = i + 1
		while not (linksEnd or i > #args) do
			thisWord = args[i]
			linksEnd = thisWord:match("%]|h|r(.-)$")
			linkName = string.format("%s %s", linkName, thisWord:match("(.-)%]|h|r") or thisWord)
			i = i + 1
		end
		-- Add final bracket and following characters "]."
		linkName = string.format("%s%s%s", linkName, linkStartAndStopString[STOP_STRING], linksEnd or "")
	else
		linkName = thisWord
		i = i + 1
		while not (linksEnd or i > #args) do
			thisWord = args[i]
			linksEnd = thisWord:match("%]|h|r")
			linkName = string.format("%s %s", linkName, thisWord)
			i = i + 1
		end
	end
		
	return linkName, linkImpossible, i
end

-------------- InterpretMessageSendToChannel()
local function InterpretMessageSendToChannel(channelArgs, args, cmdType)
	local hyperlinkValues = ""
	local messageForChannel = ""
	local flexCode
	local linkType
	local hasLinksEnd = false
	local itemLinkImpossible = true
	
	local linkEscapePrefix = GetCmdPrefix(cmdType) -- i.e.: 'auto' or 'console'
	
	assert(linkEscapePrefix)
	
	local channelNumber = channelArgs[3] -- The channel number, or nil
	
	local ammendFormatString = AMMEND_FORMAT_STRING_NO_SPACE
	local AMMEND_FORMAT_STRING_SPACE = AMMEND_FORMAT_STRING_SPACE
	
	local i = 2
	while i <= #args do -- Concatenates the full message to send, extracting raw link values
		local thisWord = args[i]
		local linkName = ""
		
		local checkLinkType = thisWord:match(linkEscapePrefix.."|c........|H(%w+):") or thisWord:match(linkEscapePrefix.."|c........|H(%w+):")
		local checkHyperlinkValues = thisWord:match(linkEscapePrefix.."|c........|H[%w]+:(.+)|h%[")
		
		if checkHyperlinkValues ~= nil and checkLinkType ~= nil and not hasLinksEnd then -- Concatenate all the words of the linkName
			linkName, itemLinkImpossible, i = SanitizeLinkGetRawValues(args, channelNumber, checkLinkType, linkEscapePrefix, i)
			linkType = checkLinkType
			hyperlinkValues = checkHyperlinkValues
			hasLinksEnd = true
		elseif checkHyperlinkValues ~= nil and checkLinkType ~= nil and hasLinksEnd then -- Just ensure extra itemLinks are in brackets, flexi can only send 1 flexiCode at a time at the moment (1.3b)
			linkName, _, i = SanitizeLinkGetRawValues(args, channelNumber, checkLinkType, linkEscapePrefix, i)
		else -- Just add this plaintext word of the message
			i = i + 1
		end
		
		messageForChannel = string.format(ammendFormatString, messageForChannel, linkName ~= "" and linkName or thisWord)
		ammendFormatString = AMMEND_FORMAT_STRING_SPACE
	end
if FLEXI.VERBOSE then print(FLEXI.flexiSpecifier..FLEXI.debugSpecifier.."linkType:"..FLEXI.Printable(linkType).." hyperlinkValues:"..FLEXI.ReadableDebugSpellOrItem(hyperlinkValues)) end

	if hyperlinkValues ~= nil and linkType == "item" then -- Item
		if itemLinkImpossible then
			flexCode = FLEXI.EncodeItemLinkToFlex(hyperlinkValues)
		end
	else -- Spell, or spell was plaintext WOW_VC or use hyperlink WOW_BCC
		local spellRank, spellId, spellName
		if FLEXI.WOW_VC then
			-- BBC and later spell links include the spellID, or something, I'm not sure really,
			-- but the spellID after '|Hspell:######' is causing a new bug in searching for names
			-- because it finds a number, not the first letter of the spell name.
			spellName = FLEXI.DeduceMostLikelySpellName(args, linkEscapePrefix) -- Bulk of the work. TODO: Optimisation target
			_, spellRank, _, _, _, _, spellId = FLEXI.GetHighestRankSpellKnown(spellName)
		else -- WOW_BCC and requires not-plaintext link (because WOW_VC would only put plaintext in the chat send box)
			if hyperlinkValues ~= nil then 
				spellId = tonumber(hyperlinkValues:match("(%d+)"))
				if type(spellId) == "number" then
					spellName = GetSpellInfo(spellId)
					spellRank = GetSpellSubtext(spellId)
				end
			end
		end
		
		if spellName ~= nil and itemLinkImpossible then
			if linkEscapePrefix ~= nil and linkEscapePrefix ~= "" then
				-- Remove linkEscapePrefix from string
				messageForChannel = messageForChannel:gsub(FLEXI.SanitizeSpecialCharacters(linkEscapePrefix..spellName), spellName)
			end
		
			local spellNameForChat = spellName
			
			linkType = "spell"
			local linkStartAndStopString = GetLinkStartAndStopString(channelNumber, linkType)
			
			if spellRank ~= nil and spellRank ~= "" then
				local spellRankPart = string.format(" (%s)", spellRank)
				spellNameForChat = string.format("%s%s", spellName, spellRankPart)
			end
			
			local spellSub = FLEXI.WOW_VC and spellName or "%["..spellName.."%]"
			
			-- TODO Will miss the pattern match if spellname contains special characters -- Rare but needs fix for future
			messageForChannel = messageForChannel:gsub(spellSub, string.format("%s%s%s", linkStartAndStopString[START_STRING], spellNameForChat, linkStartAndStopString[STOP_STRING]))
			
			if linkStartAndStopString == IMPOSSIBLE_LINK_CHANNEL_TBL then
				flexCode = FLEXI.EncodeSpellIdToFlex(spellId)
	if FLEXI.VERBOSE then print(FLEXI.flexiSpecifier..FLEXI.debugSpecifier.."spellId#"..FLEXI.Printable(spellId).." gave flexCode "..FLEXI.Printable(flexCode)) end
			end
		end
	end

	if flexCode then
		messageForChannel = string.format("%s %s", messageForChannel, flexCode)
	end

	return messageForChannel
end

-------------- CheckMessagePrintFlexi()
local function CheckMessagePrintFlexi(msgSegments)
	local fullFlexCode
	local hyperlink
	local getFlex
	local linkType

	local i = 1
	local n = 0 -- Here, n is the number of fields of the flexiCode we've found so far (12 fields for items, 1 for spells)
	while i <= #msgSegments do
		local thisWord = msgSegments[i]
		
		if n == 0 then 
			local checkFlexiLinkType = thisWord:match("flexi(.)") or thisWord:match("FLEXI(.)")
			if checkFlexiLinkType == "#" then
				linkType = FLEXI_CODE_IS_ITEM
			elseif checkFlexiLinkType == "$" then
				linkType = FLEXI_CODE_IS_SPELL
			end
			if linkType then
				getFlex = thisWord:match("flexi.(.+)") or thisWord:match("FLEXI.(.+)")
			end
		else
			getFlex = thisWord:match("(.+)")
		end
		
		if linkType == FLEXI_CODE_IS_ITEM then -- Concatenate whitespace words after finding "FLEXI#" until we have 12 fields (13 splits).
			fullFlexCode = string.format("%s%s%s", (n > 0 and fullFlexCode or ""), (n > 0 and " " or ""), getFlex) -- Prevents white-space breaking code strings, e.g.: |cff001122|Hitem:123:::...::|h[I Have Spaces]|h|r

			for _ in getFlex:gmatch("([^:]*):") do
				n = n + 1
			end
			
			if n == NUM_OF_ITEM_SPLITS then -- Valid itemLink found
				hyperlink = FLEXI.DecodeFlexToItemLink(fullFlexCode)
if FLEXI.VERBOSE == true and hyperlink ~= nil then print(FLEXI.flexiSpecifier..FLEXI.debugSpecifier.."Printing hyper-link: '"..hyperlink:gsub("|", "\\124").."'") end			
				print(FLEXI.flexiSpecifier..": "..hyperlink) -- Print the [Item Link]
				
				break
			elseif n > NUM_OF_ITEM_SPLITS then -- Invalid link found
				print(FLEXI.flexiSpecifier..FLEXI.infoSpecifier.."flexi code received had too many fields")
				
				break
			end
		elseif linkType == FLEXI_CODE_IS_SPELL then
			fullFlexCode = string.format("%s%s%s", (n > 0 and fullFlexCode or ""), (n > 0 and " " or ""), getFlex)

			for _ in getFlex:gmatch("([^:]*):") do
				n = n + 1
			end
			
			if n == NUM_OF_SPELL_SPLITS then -- Valid spellLink found
				hyperlink = FLEXI.DecodeFlexToSpellLink(fullFlexCode)
				
				if hyperlink ~= nil then
					print(FLEXI.flexiSpecifier..": "..hyperlink) -- Print the [Spell Link]
				end
				
				break
			elseif n > NUM_OF_SPELL_SPLITS then -- Invalid link found
				print(FLEXI.flexiSpecifier..FLEXI.infoSpecifier.."flexi code received had too many fields")
				
				break
			end
		end
		
		i = i + 1
	end
end

--- This function is extremely prohibitive, and will default itself to a carbon 
--- copy of the original Blizzard SubstituteChatMessageBeforeSend(msg) function 
--- if it ever throws an error, until a UI reload.
-------------- FlexiHookedSubstituteChatMessageBeforeSend()
local function FlexiHookedSubstituteChatMessageBeforeSend(msg)
if FLEXI.VERBOSE then print(FLEXI.flexiSpecifier..FLEXI.debugSpecifier.."raw message: \""..FLEXI.ReadableDebugSpellOrItem(msg).."\".") end
	if not slappedForErrorsOrPoorReturnValues then
		local args = SplitString(msg, "[^%s]+")
		local chatFrame = Chat_GetChatFrame(1)
		
		local chatSendsTo = chatFrame.editBox:GetAttribute("chatType")
		if string.upper(chatSendsTo) == "CHANNEL" then
			chatSendsTo = chatFrame.editBox:GetAttribute("channelTarget")
		end
		table.insert(args, 1, chatSendsTo)
		
		local channelArgs = {FLEXI.GetChannelArgs(args[1])}
		
		local success, flexiSubMsg = pcall(InterpretMessageSendToChannel, channelArgs, args, "auto")
		if not success then
			slappedForErrorsOrPoorReturnValues = true
			error("Flexi slapped itself. All sent chat messages will now default to standard Blizzard functionality. Msg="..FLEXI.Printable(flexiSubMsg))
		else
			return flexiSubMsg
		end
	end
	
	return msg
end

-------- FLEXI.IsSlapped()
function FLEXI.IsSlapped()
	return slappedForErrorsOrPoorReturnValues
end

---- Console cmd
SLASH_FLEXI1 = "/flexi"
SlashCmdList["FLEXI"] = function(msg)
	local args = SplitString(msg, "[^%s]+")
	
	if args == nil or args[1] == nil then
		print(FLEXI.INFO_HELPTEXT)
			
		return 1
	end
	
	local channelArgs = {FLEXI.GetChannelArgs(args[1])}
	
	if args[1]:match(".*help.*") or args[1]:match("?") then
		print(FLEXI.INFO_HELPTEXT)
			
		return 1
	elseif channelArgs[1] == nil then -- Checks again, lower, but I don't want to change arg type of 
		print(FLEXI.CHANNEL_NOT_RECOGNIZED)
		
		return -1
	end
	
	local success, messageForChannel = pcall(InterpretMessageSendToChannel, channelArgs, args, "console")
	if not success then
		print(FLEXI.flexiSpecifier..FLEXI.infoSpecifier.."An unknown problem occured while interpreting your message to the chat.")
	elseif messageForChannel ~= nil and messageForChannel ~= "" and messageForChannel ~= " " then
		SendChatMessage(messageForChannel, unpack(channelArgs)) -- Success
	end
end
--
-- /Funcs --

--- Event Funcs --
--
Frame_ChatMessage = CreateFrame("Frame", "FLEXI_ChatChannelWatcher")
Frame_ChatMessage:RegisterEvent("CHAT_MSG_CHANNEL")
Frame_ChatMessage:RegisterEvent("CHAT_MSG_SAY")
Frame_ChatMessage:RegisterEvent("CHAT_MSG_YELL")
Frame_ChatMessage:RegisterEvent("CHAT_MSG_GUILD")
Frame_ChatMessage:RegisterEvent("CHAT_MSG_OFFICER")
Frame_ChatMessage:RegisterEvent("CHAT_MSG_PARTY")
Frame_ChatMessage:RegisterEvent("CHAT_MSG_PARTY_LEADER")
Frame_ChatMessage:RegisterEvent("CHAT_MSG_RAID")
Frame_ChatMessage:RegisterEvent("CHAT_MSG_RAID_LEADER")
Frame_ChatMessage:RegisterEvent("CHAT_MSG_WHISPER_INFORM")
Frame_ChatMessage:SetScript("OnEvent",
	function(...)
		local _, _, thisChatMsg = ...
		local thisMsgSegments = SplitString(thisChatMsg, "[^%s]+")
		
		if not pcall(CheckMessagePrintFlexi, thisMsgSegments) then
			print(FLEXI.flexiSpecifier..FLEXI.infoSpecifier.."An unknown problem occured while reading a message from the chat.")
		end
	end
)

Frame_SpellBook = CreateFrame("Frame", "FLEXI_SpellBookWatcher")
Frame_SpellBook:RegisterEvent("LEARNED_SPELL_IN_TAB")
Frame_SpellBook:RegisterEvent("PLAYER_ENTERING_WORLD")
Frame_SpellBook:SetScript("OnEvent",
	function(...)
		Frame_SpellBookLoadDelay = CreateFrame("Frame", "FLEXI_SpellBookLoadDelay") -- Spellsubtexts ["Rank 5", "Racial", "Expert"] are loaded late when logging into the game world for the first time.
		Frame_SpellBookLoadDelay.limitFrames = 0.25
		Frame_SpellBookLoadDelay:SetScript("OnUpdate",
			function(_, elapsed)
				local frame = Frame_SpellBookLoadDelay
				frame.limitFrames = frame.limitFrames - elapsed
				if frame.limitFrames <= 0 then
					FLEXI.PopulateMySpellData()
					Frame_SpellBookLoadDelay:SetScript("OnUpdate", nil)
				end
			end
		)
	end
)
--
-- /Event Funcs --

-- Hooking function -- Double wrapped (see pcall'd func) and prohibitive against errors. Falls-back to blizzard implementation.
-- TODO: Create a message validity checker for the purpose of expanded opportunities for slapping oneself.
SubstituteChatMessageBeforeSend = function(msg) 
	if not slappedForErrorsOrPoorReturnValues then
		local successSub, flexiSubMsg = pcall(FlexiHookedSubstituteChatMessageBeforeSend, msg)
		local successIntegrity, integrityMaintained = pcall(IsMessageIntegrityMaintained, flexiSubMsg, msg)
		
		if string.len(flexiSubMsg) >= MAX_CHAT_MESSAGE_LEN then
			print(FLEXI.flexiSpecifier..FLEXI.infoSpecifier.."Message was too long. You can enter the chat box and press Alt+Up arrow to shorten the message and try again.")
			return msg
		end
		
		if successSub and successIntegrity and integrityMaintained == true then
if FLEXI.DEBUG then print(FLEXI.flexiSpecifier..FLEXI.debugSpecifier.."sending message string to blizzard chat function: \""..FLEXI.ReadableDebugSpellOrItem(flexiSubMsg).."\".") end
			return oldSubFunc(flexiSubMsg) -- Success
		else
			print(FLEXI.flexiSpecifier..FLEXI.debugSpecifier.."Flexi got |cffff3311SLAPPED!|r and is now blocking iteself from hooking the chat substitute. successSub="..
					FLEXI.Printable(successSub)..", successIntegrity="..FLEXI.Printable(successIntegrity)..". integrityMaintained="..
					FLEXI.Printable(integrityMaintained).."."
				) 
			slappedForErrorsOrPoorReturnValues = true -- Failed check or threw error and will never make another attempt.
		end
	end
	
	return oldSubFunc(msg)
end